/* 
 * get_head.c - the implementation of the method GET and HEAD
 * 
 * Copyright (c) 2003
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>


#include "zjhttpd.h"

/* a short introduction to this file:
 *
 * the only exported function is http_get_head()
 * It handles both, the GET and HEAD request. First
 * it tries to open the file, then the fishy things begin. If the
 * user requested ranges, the function generate_chunk_list is called
 * to create a linked list with the chunks. That is a stupid MIME like
 * multipart-stuff. This list is generated for both requests, since
 * HEAD should operate the same like GET, but without sending the body.
 *
 * The most common usage of ranges is a single range. This is optimized,
 * so there is a special function: send_range(). But since the simple
 * case doesn't need multipart-stuff it is no surprise to have this
 * functionality in a special function, is it?
 *
 * free_chunk() frees the list, doh.
 *
 * send_chunks is simple because the trick is in the generating
 * functions. send_chunk simply send every header + every body of
 * the single entries in the chunk_list{} struct.
 *
 * The most simple case is sending the whole file if no ranges are
 * requested. I'm not sure whether I should #ifdef the ragnes stuff,
 * but I think this is fast enough for normal operations, if you want
 * to have it bitchin' fast, throw the ranges-stuff away and perhaps
 * you should optimize the sending function by the usage of sendfile()
 * which implements (OS dependent!) "zero-copy", FreeBSD's manpage:
 *
 * IMPLEMENTATION NOTES
 * The FreeBSD implementation of sendfile() is "zero-copy", meaning that it
 * has been optimized so that copying of the file data is avoided.
 *           
 * but be warned, under Linux, e.g., the parameters a different and you
 * will produce unportable code.
 */

/* send the complete file to stdout */
static int send_file(int fd)
{
	int bytes;
	char buf[BUF_SIZ];

	while ((bytes = read(fd, buf, sizeof(buf))) > 0)
		fwrite(buf, 1, bytes, stdout);

	return 0;
}

/* calculate the total length of the body */
static size_t total_chunk_length(const struct chunk_list *chunks)
{
	const struct chunk_list *l;
	unsigned long total = 0;

	for (l = chunks; l; l = l->next)
	{
		total += l->header_len;
		total += l->data_len;
	}

	/* and the ending boundary --foo-- */
	total += strlen(chunks->boundary);
	total += 6;	/* heading --, trailing -- and \r\n */
		
	return total;
}


/* send the chunks - because of the good work we've done already,
 * this is only sending header, sending data and sending footer.
 */
static int send_chunks(const struct chunk_list *chunks)
{
	const struct chunk_list *l;
	size_t written;

	for (l = chunks; l; l = l->next)
	{
		written = 0;
		while (written < l->header_len)
			written += fwrite(l->header + written, 1,
					l->header_len - written, stdout);

		written = 0;
		while (written < l->data_len)
			written += fwrite(l->data + written, 1,
					l->data_len - written, stdout);
	}

	printf("--%s--\r\n", chunks->boundary);

	fflush(stdout);
	
	return 0;
}

/* free all the malloced memory in the chunk list */
static void free_chunks(struct chunk_list *chunks)
{
	struct chunk_list *p, *l = NULL;

#ifdef DEBUG
	printf("freeing chunks");
#endif

	for (p = chunks; p; p = p->next)
	{
		if (l)
		{
			free(l->header);
			free(l->data);
			free(l->boundary);
			free(l);
#ifdef DEBUG
			putchar('.');
#endif
		}
		l = p;
	}

	if (l)
	{
		free(l->header);
		free(l->data);
		free(l->boundary);
		free(l);
#ifdef DEBUG
		putchar('.');
#endif
	}

#ifdef DEBUG
	printf("\n");
#endif
}

/* read the data for a specific chunk from the file */
static int read_chunk_data(char *buf, unsigned long len, unsigned long off,
		int fd)
{
	int bytes;
	
	if (lseek(fd, off, SEEK_SET) < 0)
		return -1;

	while (len)
	{
		if ((bytes = read(fd, buf, len)) <= 0)
			return -1;
		len -= bytes;
		buf += bytes;
	}
	
	return 0;
}

/* sends the range if it is only one - I cannot use the chunk machinery
 * for that, because there is no multipart-stuff and so on, and I
 * don't want to involve all the overhead for the common case (who
 * really requests more than one range per request?)
 */
static int send_range(int fd, const struct http_ranges *r, size_t length)
{
	size_t len;
	char buf[BUF_SIZ];
	int bytes;
	
	/* calculate its data's length, +1 because 0 is not 0
	 * but the first byte, so 0-0 is 1 and 0-3 ist 4 (0, 1,
	 * 2, 3).
	 */
	len = ((r->valid & RANGE_END) ? r->end : length - 1) -
		((r->valid & RANGE_START) ? r->start : 0) + 1;

	if ((r->valid & RANGE_START) && (lseek(fd, r->start, SEEK_SET) < 0))
		return -1;

	while (len)
	{
		if ((bytes = read(fd, buf, min(len, sizeof(buf)))) <= 0)
			return -1;
		fwrite(buf, 1, bytes, stdout);
		len -= bytes;
	}
	
	return 0;
}

/* this function generates a header for the chunk. It is a separated
 * function because it is a bit Voodoo and not so easy to understand,
 * but the rest is also full of ?: and stupid casts...
 */
static void generate_chunk_header(char *header, size_t *h_len,
		const char *boundary, const char *type, size_t length,
		struct http_ranges *r)
{
	unsigned long pos = 0;
	
	pos += snprintf(header, *h_len, "--%s\r\n", boundary);

	pos += snprintf(header+pos, *h_len-pos, "Content-Type: %s\r\n", type);

	pos += snprintf(header + pos, *h_len - pos,
			"Content-Range: bytes %lu-%lu/%lu\r\n\r\n",
			(r->valid & RANGE_START) ? (unsigned long) r->start : 0,
			(r->valid & RANGE_END) ? (unsigned long) r->end : length - 1,
			(unsigned long) length);

	*h_len = pos;
}

/* FIXME: implement a memmem() here */
static int is_boundary_unique(const char *boundary, const char *data,
		const size_t data_len)
{
	/* remove the annoying warnings */
	if ((boundary) || (data) || (data_len))
		return 1;

	return 1;
}

/* builds a list of chunks from the given range */
static struct chunk_list *generate_chunk_list(const struct http_response *res,
		int fd, size_t length)
{
	struct http_ranges *r;
	char header[BUF_SIZ], boundary[BUF_SIZ];
	size_t len;
	struct chunk_list *root = NULL, *n, *last = NULL;

	/* please, please, PLEASE let this be a good random string! ;-) */
	sprintf(boundary, "%x%x%x", rand(), rand(), rand());

#ifdef DEBUG
	fprintf(stderr, "boundary is <%s>\n", boundary);
#endif

	for (r = res->ranges; r; r = r->next)
	{
#ifdef DEBUG
		fprintf(stderr, "processing next range\n");
#endif
		/* get memory for a new chunk */
		if (!(n = malloc(sizeof(*n))))
		{
			free_chunks(root);
			return NULL;
		}

		/* calculate its data's length, +1 because 0 is not 0
		 * but the first byte, so 0-0 is 1 and 0-3 ist 4 (0, 1,
		 * 2, 3).
		 */
		n->data_len = ((r->valid & RANGE_END) ? r->end : length - 1) -
			((r->valid & RANGE_START) ? r->start : 0) + 1;
#ifdef DEBUG
		printf("n->data_len = %lu\n", (unsigned long)n->data_len);
#endif

		/* +2 for \r\n */
		n->data_len += 2;

		/* get memory for the data */
		if (!(n->data = malloc(n->data_len)))
		{
			free_chunks(root);
			return NULL;
		}

#ifdef DEBUG
		fprintf(stderr, "Checkpoint 1: before read_chunk_data()\n");
#endif

		/* read the data from the file, n->data_len - 2
		 * because the two additional bytes are reserved
		 * for \r\n and do not count to the data.
		 */
		if (read_chunk_data(n->data, n->data_len - 2,
				(r->valid & RANGE_START)?r->start:0,
				fd))
		{
			free(n->data);
			free(n);
			free_chunks(root);
			return NULL;
		}

#ifdef DEBUG
		fprintf(stderr, "Checkpoint 2: before memcpy()\n");
#endif

		/* attach the \r\n */
		memcpy(n->data + n->data_len - 2, "\r\n", 2);
			
		/* check the boundary */
		if (!is_boundary_unique(boundary, n->data, n->data_len))
		{
			free(n->data);
			free(n);
			free_chunks(root);
			return NULL;
		}
		
#ifdef DEBUG
		fprintf(stderr, "Checkpoint 3: before generate_chunk_header()\n");
#endif
		
		/* generate the header for the chunk */
		len = sizeof(header);
		generate_chunk_header(header, &len, boundary,
				res->content_type, res->content_length, r);

		/* get memory for the header */
		if (!(n->header = malloc(len)))
		{
			free_chunks(root);
			return NULL;
		}

#ifdef DEBUG
		fprintf(stderr, "Checkpoint 4: before memcpy() for header\n");
#endif

		/* copy the header */
		n->header_len = len;
		memcpy(n->header, header, len);

		/* copy the boundary */
		n->boundary = strdup(boundary);

		n->next = NULL;

#ifdef DEBUG
		fprintf(stderr, "Checkpoint 5: attaching new item to list\n");
#endif

		/* attach the new chunk to the list */
		if (!root)
		{
			root = n;
			last = n;
		}
		else
		{
			last->next = n;
			last = n;
		}
	}
	
#ifdef DEBUG
	fprintf(stderr, "leaving generate_chunk_list with %p\n", root);
#endif

	return root;
}

/* check the ranges - they must not overlap or go beyond the end
 * of file. Hmm, Apache doesn't check the overlap-trick, why
 * should I do? It is really stupid and I don't want to check
 * every range against every other. Especially because I cannot
 * random access lists. Anybody a good idea for this?
 */
static int check_ranges (const struct http_ranges *ranges, const size_t max)
{
	while (ranges)
	{
		if ((ranges->valid & RANGE_END) && (ranges->end >= max))
			return 1;
		ranges = ranges -> next;
	}

	return 0;
}

/* generate a response for HEAD or GET requests */
int http_get_head(struct http_response *res)
{
	int fd, ret = HTTP_OK;
	struct chunk_list *chunks = NULL;
	
	/* first of all: check whether we really are allowed to open
	 * the file. That also prevents race conditions - if it should
	 * work, we send 200 OK and later it doesn't work... embarrassing.
	 * But if we have a open filedescriptor, our chances are better.
	 */

	fd = open(res->path, O_RDONLY);
	if (fd == -1)
	{
		if (errno == EACCES)
			return HTTP_FORBIDDEN;
		return HTTP_NOT_FOUND;
	}

	/* ok, that works... */



	/* we only need chunks if there is more than 1 range */
	if (res->ranges)
	{
		/* if ranges are requested, this is the last chance to
		 * check the proper values (e.g. no overlaping and no
		 * reading beyond end of file)
		 */

		if (check_ranges(res->ranges, res->content_length))
			return HTTP_RANGE_NOT_SATISFIABLE;

		if (res->ranges->next)
			if (!(chunks = generate_chunk_list(res, fd, res->content_length)))
				return HTTP_INTERNAL_SERVER_ERROR;
		ret = HTTP_PARTIAL_CONTENT;
	}
	
	printf("HTTP/%i.%i %i %s\r\n", res->http_major, res->http_minor,
			ret, response_message_short(ret));

	printf("Date: %s\r\n", http_datestring(time(NULL)));
	printf("Server: %s\r\n", http_serverstring());
	printf("Accept-Ranges: bytes\r\n");
	printf("Last-Modified: %s\r\n", http_datestring(res->last_modified));

	/* if there are two or more */
	if (res->ranges && res->ranges->next)
	{
		printf("Content-Type: multipart/byteranges; boundary=%s\r\n",
				chunks->boundary);
		res->content_sent = total_chunk_length(chunks);
		printf("Content-Length: %lu\r\n", (unsigned long) res->content_sent);
	}
	else
	{
		printf("Content-Type: %s\r\n", res->content_type);

		/* if one range */
		if (res->ranges)
		{
			printf("Content-Range: %lu-%lu/%lu\r\n",
					(unsigned long) ((res->ranges->valid & RANGE_START) ?
							res->ranges->start : 0),
					(unsigned long) ((res->ranges->valid & RANGE_END) ?
							res->ranges->end : res->content_length - 1),
					(unsigned long) res->content_length);
			
			res->content_sent = ((res->ranges->valid & RANGE_END) ?
					res->ranges->end : res->content_length - 1) -
					((res->ranges->valid & RANGE_START) ?
					res->ranges->start : 0);
		}
		else
		{
			/* no ranges */
			printf("Content-Length: %lu\r\n",
					(unsigned long) res->content_length);
			res->content_sent = res->content_length;
		}
	}

	printf("Connection: %s\r\n\r\n",
			(res->keep_alive == KEEP_ALIVE_YES) ? "Keep-Alive" : "Close");

	fflush(stdout);

	/* if not GET: no body.
	 * if not GET && not HEAD: CRY OUT LOUD! Who was it? Who fucked it up?
	 */
	if (res->method != METHOD_GET)
	{
		assert(res->method == METHOD_HEAD);
		close(fd);
		if (res->ranges && res->ranges->next)
			free_chunks(chunks);
		return ret;
	}

	/* only if more than one ranges */
	if (res->ranges && res->ranges->next)
	{
		if (send_chunks(chunks))
			ret = HTTP_INTERNAL_SERVER_ERROR;
		free_chunks(chunks);
	}
	else
	{
		/* only one range */
		if (res->ranges)
		{
			if (send_range(fd, res->ranges, res->content_length))
				return HTTP_INTERNAL_SERVER_ERROR;
		}
		else
		{
			if (send_file(fd))
				return HTTP_INTERNAL_SERVER_ERROR;
		}
	}

	close(fd);

	return ret;
}

