/*
 * main.c - the main function for zjhttpd
 * 
 * Copyright (c) 2003
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <syslog.h>

#include "zjhttpd.h"

/* a short introduction to this file:
 *
 * we only export main(), but this implements the whole configuration
 * stuff from commandline. The config-behaviour is: first load the
 * defaults (config_load_defaults), then read the command line arguments
 * (because there may be a -f switch to load a custom config file),
 * store the values (that's why there are so many local variables, we
 * need to "remember" the values), THEN load the user specified config
 * file (if any), if there was no such file given, load the default
 * config file (DEFAULT_CONFIG from zjhttpd.h) (this returns no error,
 * because a config file is not a MUST, we are fully operable from the
 * commandline) and THEN set the options from the command line.
 *
 * The trick is, that the command line arguments can override any
 * given option from the config file. After that we print out the
 * configuration if DEBUG is defined, switch the userid and groupid
 * (note that this is only sensible if a inetd-replacement is used,
 * that isn't able to do this itself).
 *
 * At the end we open the syslog to write our logfile entries, and
 * main_loop() from request.c is called, which does the really
 * interesting stuff.
 *
 * Later enhancements may be a private logfile (syslog isn't a really
 * good thing, I think). Perhaps a chroot-option to enhance security.
 */

static void help(const char *s)
{
	fprintf(stderr, "usage: %s [options]\n\n", s);

	fputs("-h\t\tview this help screen and exit\n", stderr);
	fputs("-f <conf-file>\tuse \"conf-file\" for reading the "
			"configuration\n", stderr);
	fputs("-u <user>\tset user (may be numeric uid or name)\n", stderr);
	fputs("-g <group>\tset group (may be numeric gid or name)\n", stderr);
	fputs("-d <dir>\tset dir as www_dir\n", stderr);
	fputs("-i <index-file>\tset index-file (e.g. \"default.htm\")\n", stderr);
	fputs("-e <error-path>\ttakes HTML error message from error-path\n",
			stderr);
	fputs("-dotfiles\tsend dotfiles if requested (default: no)\n", stderr);
	fputs("-nodotfiles\toverrides an earlier -dotfiles (from config, e.g.)\n",
			stderr);
	fputs("-ct <ct>\tuse ct as default content-type (default: text/plain)\n",
			stderr);
	fputs("-extlog\t\tturns extended logging on (+agent, +referer)\n", stderr);
	fputs("-noextlog\tturns extedned logging off\n", stderr);
	fputs("-dirlisting\tenables directory browsing\n", stderr);
	fputs("-nodirlisting\tdisables directory browsing\n", stderr);

	fputs("\nNOTE: command line arguments override options from "
			"the config file\n", stderr);
}

/* the defaults are set first, then it is checked whether a custom
 * config-file is used. This way the values that are not set in the
 * config-file are in a defined state. At the end, the additional
 * command line parameters are evaluated, so they can be used to
 * override the values from the config file. Thats why there are
 * variables for user, group, ... - if I would set them in the
 * for loop with "conf.<foo> = ", they will be overwritten when
 * the config-file is loaded.
 */
int main(int argc, char *argv[])
{
	struct server_config conf;
	int i, ret;
	const char *conf_file = NULL, *user = NULL,
		*group = NULL, *www_dir = NULL, *index = NULL, *errors = NULL,
		*content_type = NULL;
	char dotfiles = 0, extlog = 0, dirlisting = 0;

	/* we need rand() */
	srand(time(NULL) + getpid());

	/* first set the defaults */
	memset(&conf, 0, sizeof(conf));
	config_set_defaults(&conf);

	/* evaluate the additional arguments */
	for (i = 1; i < argc; i++)
	{
		/* always check whether i + 1 < argc! Why? Because it
		 * is a really bad idea to read more arguments than
		 * available. :-)
		 */

		if (!strcmp(argv[i], "-h"))
		{
			help(argv[0]);
			return 1;
		}

		/* alternative config-file */
		if (!strcmp(argv[i], "-f") && i + 1 < argc)
		{
			conf_file = argv[++i];
			continue;
		}

		/* alternative user id */
		if (!strcmp(argv[i], "-u") && i + 1 < argc)
		{
			user = argv[++i];
			continue;
		}

		/* alternative group id */
		if (!strcmp(argv[i], "-g") && i + 1 < argc)
		{
			group = argv[++i];
			continue;
		}

		/* alternative www_dir */
		if (!strcmp(argv[i], "-d") && i + 1 < argc)
		{
			www_dir = argv[++i];
			continue;
		}

		/* alternative index file */
		if (!strcmp(argv[i], "-i") && i + 1 < argc)
		{
			index = argv[++i];
			continue;
		}

		/* path for error message files */
		if (!strcmp(argv[i], "-e") && i + 1 < argc)
		{
			errors = argv[++i];
			continue;
		}

		if (!strcmp(argv[i], "-dotfiles"))
		{
			dotfiles = OPTION_ENABLE;
			continue;
		}

		if (!strcmp(argv[i], "-nodotfiles"))
		{
			dotfiles = OPTION_DISABLE;
			continue;
		}

		if (!strcmp(argv[i], "-extlog"))
		{
			extlog = OPTION_ENABLE;
			continue;
		}

		if (!strcmp(argv[i], "-noextlog"))
		{
			extlog = OPTION_DISABLE;
			continue;
		}

		if (!strcmp(argv[i], "-ct") && i + 1 < argc)
		{
			content_type = argv[++i];
			continue;
		}

		if (!strcmp(argv[i], "-dirlisting"))
		{
			dirlisting = OPTION_ENABLE;
			continue;
		}

		if (!strcmp(argv[i], "-nodirlisting"))
		{
			dirlisting = OPTION_DISABLE;
			continue;
		}

		/* huh, user told us stupid things */
		fprintf(stderr, "problem while interpreting \"%s\"\n", argv[i]);
		return 1;
	}

	/* load the configuration from a file */
	if (conf_file && (i = config_load_file(&conf, conf_file, 0)))
		return 2;

	/* if no config file was given, use the default if available,
	 * do not fail if not (we can be used with command line
	 * parameters only, no MUST for the config file).
	 */
	if (!conf_file)
		config_load_file(&conf, DEFAULT_CONFIG, 1);

	/* set the user and group IDs (and resolve names) */
	if ((user || group) && config_set_ids(&conf, user, group))
		return 3;

	/* set the new www_dir, if necessary */
	if (www_dir)
		conf.www_dir = www_dir;

	/* set the new index, ifnecessary */
	if (index)
		conf.index = index;

	/* set the error file directory */
	if (errors)
		conf.error_file_path = errors;

	/* send dotfiles? */
	switch (dotfiles)
	{
		case OPTION_ENABLE:
			conf.dotfiles = 1;
			break;
		case OPTION_DISABLE:
			conf.dotfiles = 0;
			break;
		default:
			break;
	}

	/* extended logging? */
	switch (extlog)
	{
		case OPTION_ENABLE:
			conf.extlog = 1;
			break;
		case OPTION_DISABLE:
			conf.extlog = 0;
			break;
		default:
			break;
	}
	
	/* directory browsing? */
	switch (dirlisting)
	{
		case OPTION_ENABLE:
			conf.dirlisting = 1;
			break;
		case OPTION_DISABLE:
			conf.dirlisting = 0;
			break;
		default:
			break;
	}

	/* default content-type */
	if (content_type)
		conf.default_ct = content_type;

#ifdef DEBUG
	/* print the configuration to stdout */
	config_print(&conf);
#endif

	if (conf.uid && setuid(conf.uid))
	{
		fprintf(stderr, "setuid(%u) failed: %s\n", conf.uid, strerror(errno));
		return 4;
	}
	
	if (conf.gid && setgid(conf.gid))
	{
		fprintf(stderr, "setgid(%u) failed: %s\n", conf.uid, strerror(errno));
		return 5;
	}

	/* open the syslog */
	openlog(SERVER_NAME, LOG_PID, LOG_DAEMON);
	
	/* enter the main loop */
	ret = main_loop(&conf);

	closelog();

	return ret;
}

