/*
 * misc.c - all the nice little functions you ever wanted to have :-)
 *
 * Copyright (c) 2003
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <dirent.h>

#include "zjhttpd.h"

/* a short introduction to this file:
 *
 * This is the toolbox. We export the functions:
 * cmp, get_content_type, http_datestring, http_send_dirlisting,
 * http_send_error, http_serverstring, logfile_entry, ncmp and
 * response_message_short.
 *
 * cmp is a caseless compare, ncmp the same for only n chrs.
 * 
 * http_datestring returns a RFC 1123 compatible string for the
 * given date (as time_t)
 * 
 * http_send_dirlisting generates a directory listing for a given
 * path (HTML, ugly stuff). Note that this function calls a local
 * function for putting the lines together, which may be a little
 * bit fishy, and it uses hevaily realloc, be aware of bugs, thus
 * it is tested.
 * 
 * http_send_error sends an error message (Connection: Close), it
 * determines automagically whether to send HTML or plain text.
 * 
 * http_serverstring generates a Server: line.
 *
 * logfile_entry write the entry for the logfile. Note that it needs
 * to know how many bytes have been sent, this information is stored
 * in http_request->bytes_sent, which has to be filled by the functions
 * that send bytes (some pass the value through
 * http_response->content_sent, be careful to copy the value). The
 * http_send_error function returns the number of bytes sent, so it
 * has to be catched and stored in http_request->bytes_sent.
 *
 * response_message_short (and message_long) returns a read-only string
 * matching a given HTML error code (well, not for every, only for
 * the defined).
 *
 * get_content_type tries to guess a Content-Type by examining the
 * extension of the given filename. The lookup-table is stored in
 * this file, and if no one mathes the default is returned (text/plain,
 * but can be overriden by config-file or command line).
 */

/* compares two strings, case insensitive */
int cmp(const char *a, const char *b)
{
	while (*a && (tolower(*a) == tolower(*b)))
	{
		a++;
		b++;
	}

	return *a - *b;
}

/* compares n chars of two strings, case insensitive */
int ncmp(const char *a, const char *b, size_t n)
{
	while (*a && n && (tolower(*a) == tolower(*b)))
	{
		a++;
		b++;
		n--;
	}

	return n ? (*a - *b) : 0;
}

/* return short message for an HTTP error code 
 * Note that here are short messages for codes that don't have
 * a long message. That's because this short message is more
 * useful in different places, and I think it is a good idea
 * to have all strings at one position. Note also that the
 * HTTP responses 301 and 307 need a special treatment, so they
 * cant be sent via a normal send_error() call.
 */
const char *response_message_short(int err)
{
	switch (err)
	{
		case HTTP_OK:
			return "OK";
		case HTTP_PARTIAL_CONTENT:
			return "Partial Content";
		case HTTP_MOVED_PERMANENTLY:
			return "Moved Permanently";
		case HTTP_NOT_MODIFIED:
			return "Not Modified";
		case HTTP_MOVED_TEMPORARILY:
			return "Moved Temporarily";
		case HTTP_BAD_REQUEST:
			return "Bad Request";
		case HTTP_FORBIDDEN:
			return "Forbidden";
		case HTTP_NOT_FOUND:
			return "Not Found";
		case HTTP_PRECONDITION_FAILED:
			return "Precondition Failed";
		case HTTP_RANGE_NOT_SATISFIABLE:
			return "Request Range Not Satisfiable";
		case HTTP_INTERNAL_SERVER_ERROR:
			return "Internal Server Error";
		case HTTP_NOT_IMPLEMENTED:
			return "Not Implemented";
		case HTTP_VERSION_NOT_SUPPORTED:
			return "HTTP Version Not Supported";
	}

	return "Unknown Error";
}

/* return long message for a HTTP error code */
static const char *message_long(int err)
{
	switch(err)
	{
		case HTTP_NOT_MODIFIED:
			return "This entity was not changed since your last "
				   "request. Use your localcopy.\r\n";

		case HTTP_BAD_REQUEST:
			return "Your requests looks kinda funny, are you sure "
				   "that you speak HTTP?\r\n";

		case HTTP_FORBIDDEN:
			return "You are not allowed to perform this action. "
				   "Go away.\r\n";

		case HTTP_NOT_FOUND:
			return "I could not retrieve the file or whatever you "
				   "wanted. It is really not here. Ask somewhere else.\r\n";

		case HTTP_PRECONDITION_FAILED:
			return "Your precondition failed, not sending entity.\r\n";

		case HTTP_RANGE_NOT_SATISFIABLE:
			return "I cannot send the byte range you requested.\r\n";

		case HTTP_INTERNAL_SERVER_ERROR:
			return "Oops, houston, we have a problem!\r\n";

		case HTTP_NOT_IMPLEMENTED:
			return "Funny that you are asking for that function... "
				   "uhmmm... I think it isn't implemented yet ;-)\r\n";

		case HTTP_VERSION_NOT_SUPPORTED:
			return "Howdy! That HTTP version is not supported here, "
				   "please try it somewhere else.\r\n";

	}

	return "Yo, that error has no message, please hit the programmer!\r\n";
}

/* return a RFC 1123 date string */
const char *http_datestring(const time_t t)
{
	static char buf[BUF_SIZ];

	strftime(buf, sizeof(buf), "%a, %d %b %Y %H:%M:%S GMT", gmtime(&t));

	return buf;
}

/* return our server string */
const char *http_serverstring(void)
{
	static char buf[BUF_SIZ];

	sprintf(buf, "zjhttpd/%i.%i-%s", VERSION_MAJOR, VERSION_MINOR,
			VERSION_STRING);

	return buf;
}

/* send the error message, plain text */
static unsigned long  send_plaintext_error(const struct http_request *req,
		int err)
{
	const char *message;
	unsigned long length;
	
	/* if the error occured in header parsing, http_major and http_minor
	 * are not set, or nonsens is set. After sanity_check the value
	 * http_version_checked is 1 meaning the content is ok.
	 */
	if (req->http_version_checked)
		printf("HTTP/%i.%i %i %s\r\n", req->http_major, req->http_minor, err,
			response_message_short(err));
	else
		printf("HTTP/%i.%i %i %s\r\n", HTTP_VERSION_MAJOR, HTTP_VERSION_MINOR,
			err, response_message_short(err));

	message = message_long(err);
	length = strlen(message);

	printf("Date: %s\r\n", http_datestring(time(NULL)));
	printf("Server: %s\r\n", http_serverstring());
	printf("Content-Type: text/plain\r\n");
	printf("Content-Length: %lu\r\n", length);
	printf("Connection: Close\r\n\r\n");

	printf("%s", message);

	fflush(stdout);

	return length;
}

/* send the HTML error message */
static unsigned long  send_html_error(struct http_request *req, int err, int fd)
{
	struct stat stbuf;
	char buf[BUF_SIZ];
	int bytes;

	fstat(fd, &stbuf);
	
	/* if the error occured in header parsing, http_major and http_minor
	 * are not set, or nonsens is set. After sanity_check the value
	 * http_version_checked is 1 meaning the content is ok.
	 */
	if (req->http_version_checked)
		printf("HTTP/%i.%i %i %s\r\n", req->http_major, req->http_minor, err,
			response_message_short(err));
	else
		printf("HTTP/%i.%i %i %s\r\n", HTTP_VERSION_MAJOR, HTTP_VERSION_MINOR,
			err, response_message_short(err));
	
	printf("Date: %s\r\n", http_datestring(time(NULL)));
	printf("Server: %s\r\n", http_serverstring());
	printf("Content-Type: text/html\r\n");
	printf("Content-Length: %lu\r\n", (unsigned long)stbuf.st_size);
	printf("Connection: Close\r\n\r\n");

	while((bytes = read(fd, buf, sizeof(buf))) > 0)
		fwrite(buf, 1, bytes, stdout);
	
	fflush(stdout);

	return stbuf.st_size;
}

/* send the error message - uses HTML if available, otherweise plain text */
unsigned long http_send_error(const struct server_config *conf,
		struct http_request *req, int err)
{
	char buf[BUF_SIZ];
	int fd, ret;

	/* clear keep-alive when Connection: Close is sent. */
	req->keep_alive = KEEP_ALIVE_NONE;

	if (conf->error_file_path)
	{
		snprintf(buf, sizeof(buf), "%s/%i.html", conf->error_file_path, err);
		fd = open(buf, O_RDONLY);
		if (fd == -1)
			return send_plaintext_error(req, err);
		ret = send_html_error(req, err, fd);
		close(fd);
		return ret;
	}

	return send_plaintext_error(req, err);
}

static const char* content_type_table[][2] = {
	{ "txt", "text/plain" },
	{ "asc", "text/plain" },
	{ "html", "text/html" },
	{ "htm", "text/html" },
	{ "doc", "application/msword" },	/* ARGH! ;-) */
	{ "pdf", "application/pdf" },
	{ "ai", "application/postscript" },
	{ "eps", "application/postscript" },
	{ "ps", "application/postscript" },
	{ "rtf", "application/rtf" },
	{ "cpio", "application/x-cpio" },
	{ "csh", "application/x-csh" },
	{ "dvi", "application/x-dvi" },
	{ "gtar", "application/x-gtar" },
	{ "latex", "application/x-latex" },
	{ "sh", "application/x-sh" },
	{ "shar", "application/x-shar" },
	{ "tar", "application/x-tar" },
	{ "tcl", "application/x-tcl" },
	{ "tex", "application/x-tex" },
	{ "man", "application/x-troff-man" },
	{ "zip", "application/zip" },
	{ "au", "audio/basic" },
	{ "snd", "audio/basic" },
	{ "wav", "audio/x-wav" },
	{ "gif", "image/gif" },
	{ "ief", "image/ief" },
	{ "jpeg", "image/jpeg" },
	{ "jpg", "image/jpeg" },
	{ "jpe", "image/jpeg" },
	{ "png", "image/png" },
	{ "tiff", "image/tiff" },
	{ "tif", "image/tiff" },
	{ "xbm", "iamge/x-xbitmap" },
	{ "xpm", "image/x-xpixmap" },
	{ "rtx", "text/richtext" },
	{ "xml", "text/xml" },
	{ "dtd", "text/xml" },
	{ "mpeg", "video/mpeg" },
	{ "mpg", "video/mpeg" },
	{ "mpe", "video/mpeg" },
	{ "qt", "video/quicktime" },
	{ "mov", "video/quicktime" },
	{ "avi", "video/x-msvideo" },
	{ NULL, NULL }		/* KEEP THIS! */
};

/* retrieve the Content-Type for a given file extension */
const char *get_content_type(const char *path, const struct server_config *conf)
{
	char *p;
	const char *ret = conf->default_ct;
	int i = 0;
	
	p = strrchr(path, '.');
	if (!p || !*p || !*(p + 1))
		return NULL;
	p++;

	while (content_type_table[i][0])
	{
		if (!cmp(content_type_table[i][0], p))
			return content_type_table[i][1];
		i++;
	}

	return ret;
}

int logfile_entry(const struct server_config *conf,
		const struct http_request *req, int code)
{
	struct sockaddr_in in;
	struct tm *tm;
	time_t now;
	char date[BUF_SIZ];
	socklen_t in_size = sizeof(in);

	getsockname(STDIN_FILENO, (struct sockaddr*)&in, &in_size);

	now = time(NULL);
	tm = localtime(&now);

	strftime(date, sizeof(date), "%d/%b/%Y:%H:%M:%S %z", tm);

	if (conf->extlog)
		syslog(LOG_INFO, "%s - - [%s] \"%s\" %i %lu <%s> <%s>\n",
				inet_ntoa(in.sin_addr), date, req->request_line, code,
				(unsigned long) req->bytes_sent,
				(req->referer)?(req->referer):("no referer"),
				(req->agent)?(req->agent):("no agent"));
	else
		syslog(LOG_INFO, "%s - - [%s] \"%s\" %i %lu\n", inet_ntoa(in.sin_addr),
				date, req->request_line, code,
				(unsigned long) req->bytes_sent);
	
	return 0;
}

/* generate a single line in the listing
 * This function is also a good buffer overflow candidate
 */
static int generate_dirent_line(char *buf, unsigned long buf_len,
		const char *filename, const char *datestring, unsigned long size,
		int is_dir)
{
	unsigned long len = 0;
	int i, unit = 0;
	double real_size;

	*buf = '\0';

	/* first: two spaces */
	if (len + 3 < buf_len)
	{
		strcpy(buf, "  ");
		len += 2;
		buf += 2;
	}

	/* now the <a href="file"> stuff */
	if (len + 13 + strlen(filename) < buf_len)
	{
		strcpy(buf, "<a href=\"");
		len += 9;
		buf += 9;
		strcpy(buf, filename);
		len += strlen(filename);
		buf += strlen(filename);
		strcpy(buf, "\">");
		len += 2;
		buf += 2;
	}

	/* now the filename, max 50 chars, fill up with ' ' */
	if (len + 55 < buf_len)
	{
		for (i = 0; (i < 50) && filename[i]; i++, len++)
			*(buf++) = filename[i];

		strcpy(buf, "</a>");
		buf += 4;
		len += 4;

		while (i++ < 50)
		{
			*(buf++) = ' ';
			len++;
		}
	}

	/* add two spaces and then the date string */
	if (len + 3 + strlen(datestring) < buf_len)
	{
		i = sprintf(buf, "  %s", datestring);
		len += i;
		buf += i;
	}

	if (is_dir)
	{
		if (len + 3 < buf_len)
		{
			strcpy(buf, "\r\n");
			len += 2;
		}

		return len;
	}

	/* now add five spaces */
	if (len + 6 < buf_len)
	{
		strcpy(buf, "     ");
		buf += 5;
		len += 5;
	}

	real_size = size;
	while (real_size > 1024)
	{
		unit++;
		real_size /= 1024;
	}
	
	if (unit)
		i = snprintf(buf, buf_len - len, "%6.1f ", real_size);
	else
		i = snprintf(buf, buf_len - len, "%6lu ", size);

	buf += i;
	len += i;

	if (len + 5 < buf_len)
	{
		len++;
		switch(unit)
		{
			default:
			case 0:
				*(buf++) = ' ';
				break;

			case 1:
				*(buf++) = 'K';
				break;

			case 2:
				*(buf++) = 'M';
				break;

			case 3:
				*(buf++) = 'G';
				break;

			case 4:
				*(buf++) = 'T';
				break;
		}

		strcpy(buf, "B\r\n");
		len += 3;
	}
		
	return len;
}

/* generate the directory listing for the given path
 * FIXME: check this function good for buffer overflows,
 * it is the best candidate, I think.
 */
static int generate_dirlisting(const char *path, const char *url_path,
		char **dir_buf, unsigned long *dir_len)
{
	unsigned long buf_used = 0, buf_len = BUF_SIZ, len;
	char *buf, linebuf[BUF_SIZ], namebuf[BUF_SIZ], timebuf[BUF_SIZ];
	DIR *dir;
	struct dirent *dirent;
	struct stat stbuf;
	struct tm *tm;

	if(!(dir = opendir(path)))
		return 1;

	if (!(buf = malloc(buf_len)))
		return 2;

	/* looks like Voodoo, but is worse: HTMLish Design!!!!111 */
	buf_used = snprintf(buf, buf_len,
			"<html>\r\n"
			"<head>\r\n"
			"  <title>Index of %s</title>\r\n"
			"</head>\r\n"
			"<body>\r\n"
			"<h1>Index of %s</h1>\r\n"
			"<pre>  Name                                                 "
			"     Last Modified          Size\r\n\r\n\r\n",
			url_path, url_path);

	while ((dirent = readdir(dir)))
	{
		/* NOTE: stat_file appends a / to the dir, so %s%s works */
		snprintf(namebuf, sizeof(namebuf), "%s%s", path, dirent->d_name);
		if (stat(namebuf, &stbuf))
			continue;

		tm = localtime(&stbuf.st_mtime);
		strftime(timebuf, sizeof(timebuf), "%H:%M:%S %F", tm);

		len = generate_dirent_line(linebuf, sizeof(linebuf), dirent->d_name,
				timebuf, stbuf.st_size,
				((stbuf.st_mode & S_IFDIR) == S_IFDIR));

		if (buf_used + len >= buf_len)
		{
			buf_len += BUF_SIZ;
			if(!(buf = realloc(buf, buf_len)))
				return HTTP_INTERNAL_SERVER_ERROR;
		}

		memcpy(buf + buf_used, linebuf, len);
		buf_used += len;
	}

	closedir(dir);

	len = snprintf(linebuf, sizeof(linebuf),
			"</pre>\r\n</body>\r\n</html>\r\n");

	if (buf_used + len >= buf_len)
	{
		buf_len += BUF_SIZ;
		if(!(buf = realloc(buf, buf_len)))
			return HTTP_INTERNAL_SERVER_ERROR;
	}

	memcpy(buf + buf_used, linebuf, len);
	buf_used += len;

	*dir_buf = buf;
	*dir_len = buf_used;

	return 0;
}

/* send a directory listing */
int http_send_dirlisting(const struct http_request *req,
		struct http_response *res)
{
	unsigned long dir_len;
	char *dir_buf;

	if (generate_dirlisting(res->path, req->path, &dir_buf, &dir_len))
		return HTTP_INTERNAL_SERVER_ERROR;
	
	printf("HTTP/%i.%i %i %s\r\n", res->http_major, res->http_minor,
			HTTP_OK, response_message_short(HTTP_OK));

	printf("Date: %s\r\n", http_datestring(time(NULL)));
	printf("Server: %s\r\n", http_serverstring());
	printf("Accept-Ranges: bytes\r\n");
	printf("Content-Length: %lu\n", dir_len);
	printf("Content-Type: text/html\r\n");
	printf("Connection: %s\r\n\r\n",
			(res->keep_alive == KEEP_ALIVE_YES) ? "Keep-Alive" : "Close");

	fflush(stdout);
	fwrite(dir_buf, 1, dir_len, stdout);

	free(dir_buf);

	res->content_sent = dir_len;
	
	return HTTP_OK;
}
