/*
 * request.c - reading and parsing the requests
 * 
 * Copyright (c) 2003
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>

#include "zjhttpd.h"

static int out_of_memory;

/* a short introduction to this file:
 *
 * This is te really tricky thing!
 *
 * strategy for avoiding buffer overflows: use BUF_SIZ for every
 * buffer - since line is a char[BUF_SIZ], it cannot be too long
 * for copying it completely into the local buffer, and in most
 * cases we copy only a part of it.
 *
 * entry point is main_loop (read the long comment there, too).
 * The lifetime cycle of a request:
 *
 * - read it from stdin (read_request_from_stdin)
 * - do sanity checks (sanity_checks)
 * - if DEBUG: print out the request
 * - if Keep-Alive is not requested, set a flag to 0 to leave
 *   the main loop after the first (and in this case last) request
 * - send response (send_response)
 * - free the allocated memory (free_request)
 * - loop until Keep-Alive isn't requested any more
 *
 * Note: on errors the loop is left earlier, because all error messages
 * are using Connection: Close.
 *
 * Once a request is read from stdin, it is parsed. When the first line
 * is read, parse_method_line is called, the rest will be checked with
 * ncmp for Host:, Content-Type: and so on.
 *
 * parse_method_line stores the interesting parts (URI, path, method, ...)
 * and examines the HTTP version (local function get_http_version).
 * Note that the version is only stored, not tested, that is done later
 * in sanity_checks.
 *
 * The different header-lines are parsed by special functions, like
 * Host: -> get_header_host
 * but the generic ones are only copied (strdup, because the whole
 * header is in a local buffer, so we need to "rescue" the interesting
 * values) with get_header_value.
 * Content-Length: -> get_content_length
 * the If-Modified-Since and If-Unmodified-Since are interesting,
 * because they need date parsing. For that get_mod_date is called,
 * which may call get_rfc_1036_date or get_asctime_date if an old
 * date string is used (HTTP/1.1 tells the user not to send this
 * crap).
 * Connection: -> get_keep_alive_state
 *
 * A more interesting part is the range-stuff. Users can request
 * only parts of the files. The simple way is only one range, but
 * they can request many ranges (e.g. Range: bytes=1-10,20-25,30-).
 * We create a linked list with the ranges, each entry has a start,
 * a end and a flag that indicades which members are valid.
 *
 * Expect-Headers create a HTTP_NOT_IMPLEMENTED error.
 *
 *
 * Well, this file is long, and it is tricky. It is perhaps the most
 * tricky part at all, since all the parsing is done here. I want to
 * remember: my last httpd has some serious bugs, that were caused
 * by buffer overflows, most of them in the parsing part. I try to
 * solve this problem by using HUGE buffers for every thing (and,
 * of course, the same value for HUGE), so there is no possibility
 * to overflow, I think.
 */

/* this function gets something like "http://www.foo:80/bla/gnarf"
 * and fills in the struct-fields uri, uri_host, port, path
 * and uri_data. The latter is a pointer into uri at the position
 * where the data starts.
 */
static void parse_uri(struct http_request *req, const char *uri)
{
	char host[BUF_SIZ];
	char port[BUF_SIZ];
	char path[BUF_SIZ];
	char *p, *start;
	int path_is_set = 0, host_is_set = 0, port_is_set = 0;
	unsigned int uri_data_offset = 0;

	/* save a pointer to the start for calculating the offset.
	 * I have to discard the const, sorry. But I know that
	 * I don't write into the string, okay? :-)
	 */
	start = (char*)uri;

	/* if there is a "real" URL we should be able to deal with this.
	 * it is really braindead, but RFC2616 tells us to do so.
	 */
	if (!ncmp(uri, "http://", 7))
	{
		uri += 7;
		p = host;

		while (*uri && (*uri != '/') && (*uri != ':'))
			*(p++) = *(uri++);

		*p = '\0';

		host_is_set = 1;

		/* host now contains the host-part of the uri, if there
		 * is one. Now uri points to either :, / or \0.
		 */

		if (*uri == ':')
		{
			uri++;
			p = port;

			while (*uri && isdigit(*uri))
				*(p++) = *(uri++);

			*p = '\0';

			/* port now contains the port if there is one, and
			 * uri points either to / or to \0.
			 */

			port_is_set = 1;
		}

		/* something like "http://www.foo.bar", not nice. */
		if (!*uri)
		{
			strcpy(path, "/");
			path_is_set = 1;
		}
	}

	if (!path_is_set)
	{
		p = path;
		while (*uri && (*uri != '?'))
			*(p++) = *(uri++);
		*p = '\0';
	}

	/* path now contains a path (either '/' or something useful),
	 * and uri points to '?' if there is a parameter, or '\0'
	 * if not. If there is some data, we save the offset and
	 * adjust the pointers. Why? Haha, that is a nice pitfall.
	 * We have to copy all the string-stuff via strdup() because
	 * this is a local buffer. If we copy it, we have new
	 * addresses and if we save only the pointer into the old
	 * string and not the offset for adjusting the new pointer,
	 * ouch ;-)
	 */

	if (*uri == '?')
		uri_data_offset = (uri - start) + 1;

	/* ok, everything is in its place, just copy it to the
	 * final place via strdup(). Only strdup the stuff that
	 * exists! It is easier to handle a NULL than to check
	 * the address instead of the derefereced pointer.
	 */

	req->uri = strdup(start);

	if (uri_data_offset)
		req->uri_data = req->uri + uri_data_offset;

	if (host_is_set)
		req->uri_host = strdup(host);

	if (port_is_set)
		req->uri_port = strtoul(port, NULL, 0);
	
	req->path = strdup(path);
}


/* this function fills in the http_major and http_minor fields of
 * the struct.
 */
static int get_http_version(struct http_request *req, const char *line)
{
	char major[BUF_SIZ], minor[BUF_SIZ], *p;

	p = major;
	
	while (*line && isdigit(*line))
		*(p++) = *(line++);

	*p = '\0';

	/* major now contains the major version, line points to the '.' */

	if (!*line || *line != '.')
		return 1;

	line++;

	/* unexpected end? */
	if (!*line)
		return 1;

	p = minor;

	while (*line && isdigit(*line))
		*(p++) = *(line++);

	*p = '\0';

	/* minor now contains the minor version, line points to
	 * the first nondigit. If this is not a whitespace, the
	 * the user told us bullshit and we should tell him.
	 */

	if (!*line || !isspace(*line))
		return 1;

	/* now convert the strings major[] and minor[] into the
	 * right values, unsigned int.
	 */

	req->http_major = strtoul(major, NULL, 0);
	req->http_minor = strtoul(minor, NULL, 0);

	return 0;
}

/* this function gets the first line of the request and has much work
 * to do :-)
 */
static int parse_method_line(struct http_request *req, const char *line)
{
	char method[BUF_SIZ], *p, uri[BUF_SIZ];

	p = method;
	
	while (*line && !isspace(*line))
		*(p++) = *(line++);

	*p = '\0';

	/* method now contains GET, HEAD, POST or anything else, line
	 * points to the first whitespace after the method.
	 * Note that METHOD_NONE is zero, so it is set by memset(3).
	 */

	if (!cmp(method, "get"))
		req->method = METHOD_GET;
	if (!cmp(method, "head"))
		req->method = METHOD_HEAD;

	/* NOTE: POST will definetly be implemented, but not in 0.1.0
	if (!cmp(method, "post"))
		req->method = METHOD_POST;
*/
	if (req->method == METHOD_NONE)
		return HTTP_NOT_IMPLEMENTED;

	/* skip the whitespaces between METHOD and URI */
	while (*line && isspace(*line))
		line++;

	/* unexpected end of request? */
	if (!*line)
		return HTTP_BAD_REQUEST;

	p = uri;

	while (*line && !isspace(*line))
		*(p++) = *(line++);

	*p = '\0';

	/* unexpected end of request? */
	if (!*line)
		return HTTP_BAD_REQUEST;

	/* uri now contains the URI, line points to the first
	 * whitespace after the URI. Next to read: HTTP/...
	 */
	parse_uri(req, uri);

	/* skip whitespaces between URI and HTTP */
	while (*line && isspace(*line))
		line++;

	/* unexpected end of request? */
	if (!*line)
		return HTTP_BAD_REQUEST;

	/* huh, client does not speak HTTP? */
	if (ncmp(line, "http/", 5))
		return HTTP_BAD_REQUEST;

	/* skip "HTTP/" */
	line += 5;

	/* unexpected end of request? */
	if (!*line)
		return HTTP_BAD_REQUEST;

	if (get_http_version(req, line))
		return HTTP_VERSION_NOT_SUPPORTED;

	/* skip the HTTP version */

	while (*line && !isspace(*line))
		line++;

	/* skip the whitespaces after HTTP/n.n */

	while (*line && isspace(*line))
		line++;

	/* so, we should now be at the end of the line,
	 * because isspace() is true for spaces and for
	 * CR and LF - so if we find here something that
	 * is not \0 the user appended some shit and we
	 * should tell him about that.
	 */
	
	if (*line)
		return HTTP_BAD_REQUEST;
	
	return 0;
}

/* this function gets the value of a header (-> the first
 * non-whitespace after the first ':') and strdup(3)s it,
 * but cuts te \r\n.
 */
static char *get_header_value(const char *line)
{
	char buf[BUF_SIZ], *p;

	while (*line && (*line != ':'))
		line++;
	if (*line)
		line++;
	while (*line && isspace(*line))
		line++;

	p = buf;
	while (*line && (*line != '\r') && (*line != '\n'))
		*(p++) = *(line++);
	*p = '\0';

	if (buf[0] == '\0')
		return NULL;
	
	return strdup(buf);
}

/* gets the content-length from the header */
static size_t get_content_length(const char *line)
{
	char buf[BUF_SIZ];	/* no overflow: line is not greater than BUF_SIZ */
	char *p = buf;
	
	while (*line && !isdigit(*line))
		line++;

	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	return strtoul(buf, NULL, 0);
}

static const char *month_table[] = {"jan", "feb", "mar", "apr", "may",
	"jun", "jul", "aug", "sep", "oct", "nov", "dec"};

/* helper function: gets the index of a month (0 jan, 1 feb, ...) */
static int get_month_index(const char *month)
{
	int i;

	for (i = 0; i < 12; i++)
		if (!cmp(month, month_table[i]))
			return i;

	return -1;
}

/* gets the date from a RFC-1036-Datestring.
 * Note: we don't start at the beginning of the header line,
 * but we are currently on the comma after the long weekday
 */
static time_t get_rfc_1036_date(const char *line)
{
	struct tm tm;
	char buf[BUF_SIZ], *p = buf;

	memset(&tm, 0, sizeof(tm));

	while (*line && !isdigit(*line))
		line++;

	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the day of the month, line
	 * is on the '-' between day of month and month name
	 */

	tm.tm_mday = atoi(buf);

	/* unexpected end of line? */
	if (!*line)
		return 0;
	
	if (!*(++line))
		return 0;

	/* we are on the first char of the (short) month name */

	p = buf;
	while (*line && (*line != '-'))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the short month name, line points to
	 * the '-' between month name and year.
	 */
	if ((tm.tm_mon = get_month_index(buf)) == -1)
		return 0;

	/* unexpected end of line? */
	if (!*line)
		return 0;
	
	if (!*(++line))
		return 0;

	p = buf;
	while (*line && !isspace(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the year without century,
	 * line points to the first whitespace between
	 * year and hour.
	 */
	tm.tm_year = atoi(buf);

	while (*line && isspace(*line))
		line++;

	/* unexpected end of line? */
	if (!*line)
		return 0;

	/* line now points to the first digit of the hour */

	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the hour, line points to the ':'
	 * between hour and minute.
	 */

	tm.tm_hour = atoi(buf);

	/* unexepcted end of line? */
	if (!*line)
		return 0;
	
	if (!*(++line))
		return 0;

	/* line now points to the first digit of the minute */
	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the minute, line points to the ':'
	 * between minute and second.
	 */

	tm.tm_min = atoi(buf);

	/* unexepcted end of line? */
	if (!*line)
		return 0;
	
	if (!*(++line))
		return 0;

	/* line now points to the first digit of the second */
	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the second, line points to the first
	 * whitespace between seconds and timezone
	 */

	tm.tm_sec = atoi(buf);

	while (*line && isspace(*line))
		line++;

	/* unexpected end of line? */
	if (!*line)
		return 0;

	/* line now points to GMT, we hope */
	p = buf;
	while (*line && !isspace(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the timezone */

	tm.tm_zone = buf;
	
	return timegm(&tm);
}

/* get the date from a asctime-Datestring.
 * Note: we don't start at the beginning of the header line,
 * but we are currently on the first whitespace between
 * the short weekday and the monthname.
 */
static time_t get_asctime_date(const char *line)
{
	struct tm tm;
	char buf[BUF_SIZ], *p = buf;

	memset(&tm, 0, sizeof(tm));

	while (*line && isspace(*line))
		line++;

	/* line now points to the first char of the month name */
	p = buf;
	while (*line && !isspace(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the month name, line points to the
	 * first whitespace between month name and day of month
	 */
	if ((tm.tm_mon = get_month_index(buf)) == -1)
		return 0;

	while (*line && isspace(*line))
		line++;

	/* line now points to the first digit of the day of month */
	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the day of month, line points to the
	 * first whitespace between day of month and hour.
	 */

	tm.tm_mday = atoi(buf);

	while (*line && isspace(*line))
		line++;

	/* unexepcted end of line? */
	if (!*line)
		return 0;

	/* line now points to the hour */
	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the hour, line points to the ':'
	 * between hour and minute.
	 */

	tm.tm_hour = atoi(buf);

	/* unexepcted end of line? */
	if (!*line)
		return 0;
	
	if (!*(++line))
		return 0;

	/* line now points to the first digit of the minute */
	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the minute, line points to the ':'
	 * between minute and second.
	 */

	tm.tm_min = atoi(buf);

	/* unexepcted end of line? */
	if (!*line)
		return 0;
	
	if (!*(++line))
		return 0;

	/* line now points to the first digit of the second */
	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the second, line points to the first
	 * whitespace between seconds and year
	 */

	tm.tm_sec = atoi(buf);

	while (*line && isspace(*line))
		line++;

	/* line now points to the year (with century!) */
	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	tm.tm_year = atoi(buf) - 1900;

	/* we assume GMT */

	return timegm(&tm);
}

/* gets the modified-date from the header.
 * we parse something like "If-Modified-Since: Tue, 05 May 1996 10:23:10 GMT"
 *
 * NOTE: HTTP/1.1 requires the ability to read all three time formats.
 * The general from RFC 1123 is shown above, that is the official one that
 * SHOULD (or MUST?) be used by HTTP/1.1 software, but there is still a
 * lot of broken stuff out there. The other two are:
 *
 * Monday, 06-May-96 04:57:00 GMT (RFC 1036)
 * Mon May 6 04:57:00 1996 (ANSI-C-asctime() format)
 */
static time_t get_mod_date(const char *line)
{
	struct tm tm;
	char buf[BUF_SIZ], *p = buf;

	memset(&tm, 0, sizeof(tm));

	while (*line && (*line != ':'))
		line++;

	/* unexpected end of line? */
	if (!*line)
		return 0;

	if (!*(++line))
		return 0;

	while (*line && isspace(*line))
		line++;

	/* now we are on the first char of the weekday */
	p = buf;
	while (*line && !isspace(*line) && (*line != ','))
		*(p++) = *(line++);
	*p = '\0';

	/* we can now determine which date format is given:
	 * RFC-1123 has a short weekday followed by a comma,
	 * RFC-1036 has a long weekday without a comma
	 * asctime() has a short weekday without a comma
	 * So: if *line is now a comma, we have the common case,
	 * the HTTP/1.1 conform RFC-1123 date. We will do the
	 * rest in this function (as said, the common case, and
	 * thus we optimize for it).
	 * If strlen(buf) is larger than 3 (which is the length
	 * of a short weekday), we have the RFC-1036 format,
	 * the else tree is asctime().
	 */

	if (*line != ',')
	{
#ifdef DEBUG
		printf("Aha, it's an asctime date!\n");
#endif
		return get_asctime_date(line);
	}

	if (strlen(buf) > 3)
	{
#ifdef DEBUG
		printf("Aha, it's a RFC 1036 date!\n");
#endif
		return get_rfc_1036_date(line);
	}

#ifdef DEBUG
	printf("Aha, it's a RFC 1123 date!\n");
#endif

	/* so it is a RFC 1123 date */

	/* now we are on the first whitespace OR comma after the weekday */
	while (*line && !isdigit(*line))
		line++;
	/* line points now to the first digit of the day */
	
	/* unexpected end of line */
	if (!*line)
		return 0;

	p = buf;

	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the day of the month, line points
	 * to the first non-digit, the whitespace between the
	 * day of the month and the month.
	 */

	tm.tm_mday  = atoi(buf);

	while (*line && isspace(*line))
		line++;

	/* unexpected end of line */
	if (!*line)
		return 0;

	/* line now points to the first char of the month */

	p = buf;
	while (*line && !isspace(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the month, line points to the first
	 * whitespace, between month and year
	 */

	if ((tm.tm_mon = get_month_index(buf)) == -1)
		return 0;

	while (*line && !isdigit(*line))
		line++;

	/* unexpected end of line */
	if (!*line)
		return 0;

	/* line now points to the year */

	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf contains now the year, line points to the first
	 * non-digit, the whitespace between year and hour
	 */

	tm.tm_year = atoi(buf) - 1900;

	while (*line && !isdigit(*line))
		line++;

	/* unexpected end of line */

	if (!*line)
		return 0;

	/* line now points to the hour of the day */

	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf contains the hour of the day */

	tm.tm_hour = atoi(buf);

	/* something strange, but not a date string */
	if (*line != ':')
		return 0;
	line++;

	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the minutes */

	tm.tm_min = atoi(buf);

	/* something strange, but not a date string */
	if (*line != ':')
		return 0;
	line++;

	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the seconds */

	tm.tm_sec = atoi(buf);

	while (*line && isspace(*line))
		line++;

	/* unexpected end of line */
	if (!*line)
		return 0;

	p = buf;
	while (*line && !isspace(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the timezone */

	tm.tm_zone = buf;

	return timegm(&tm);
}

/* gets the kepp-alive-state from the header entry */
static enum keep_alive_states get_keep_alive_state(const char *line)
{
	while (*line && (*line != ':'))
		line++;
	if (*line)
		line++;
	while (*line && isspace(*line))
		line++;

	if (!ncmp(line, "close", 5))
		return KEEP_ALIVE_NO;

	if (!ncmp(line, "keep-alive", 10))
		return KEEP_ALIVE_YES;

	return KEEP_ALIVE_NONE;
}

/* free ranges list */
static void free_ranges(struct http_ranges *r)
{
	struct http_ranges *p, *l = NULL;

	for (p = r; p; p = p->next)
	{
		if (l)
			free(l);
		l = p;
	}

	if (l)
		free(l);
}

/* gets the ranges from the header */
static struct http_ranges *get_ranges(const char *line)
{
	char buf[BUF_SIZ], *p;
	struct http_ranges *root = NULL, *n, *l;

	/* skip Range: */
	while (*line && (*line != ':'))
		line++;
	if (*line)
		line++;

	/* skip whitespaces after ':' */
	while (*line && isspace(*line))
		line++;

	/* unexpected end of line? */
	if (!*line)
		return NULL;

	p = buf;
	while (*line && !isspace(*line) && (*line != '='))
		*(p++) = *(line++);
	*p ='\0';

	/* buf MUST contain now "bytes", otherwise the client is fooling us */
	if (cmp(buf, "bytes"))
		return NULL;

	while (*line && !(isdigit(*line) || (*line == '-')))
		line++;

	/* unexpected end of line? */
	if (!*line)
		return NULL;

	/* line now points to the first digit of the first range or '-' */

	for(;;)
	{
		/* new range */
		n = calloc(1, sizeof(*n));
		if (!n)
		{
			out_of_memory = 1;
			return NULL;
		}

		/* does it have a start? */
		if (*line != '-')
		{
			p = buf;
			while (*line && isdigit(*line))
				*(p++) = *(line++);
			*p = '\0';
			n->valid |= RANGE_START;
			n->start = strtoul(buf, NULL, 0);
		}

		/* is it a valid range? */
		if (*line != '-')
		{
			free(n);
			break;
		}

		line++;

		/* does it have an end? */
		if (*line && !isspace(*line))
		{
			p = buf;
			while (*line && isdigit(*line))
				*(p++) = *(line++);
			*p = '\0';
			n->valid |= RANGE_END;
			n->end = strtoul(buf, NULL, 0);
		}

		/* short sanity check. RFC 2616 says: if the start is greater
		 * than the end, the user told us bullshit, so we MUST ignore
		 * the whole ranges-stuff.
		 */
		if (n->valid == (RANGE_END | RANGE_START))
		{
			if (n->start > n->end)
			{
				free_ranges(root);
				free(n);
				return NULL;
			}
		}

		/* attach new range at the end of the list */
		if (root)
		{
			for (l = root; l->next; l = l->next)
				;
			l->next = n;
		}
		else	/* no list? make new! */
			root = n;

		/* if there is a ',', it will be followed by the next range */
		if (*line != ',')
			break;

		line++;
	}

	return root;
}

static void get_header_host(struct http_request *req, const char *line)
{
	char buf[BUF_SIZ], *p;

	/* skip 'Host:' */
	while (*line && (*line != ':'))
		line++;
	if (*line)
		line++;
	while (*line && isspace(*line))
		line++;

	/* unexpected end of line? */
	if (!*line)
		return;

	p = buf;

	while (*line && !isspace(*line) && (*line != ':'))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the hostname */
	req->header_host = strdup(buf);

	/* only host is given */
	if (*line != ':')
		return;
	line++;

	p = buf;
	while (*line && isdigit(*line))
		*(p++) = *(line++);
	*p = '\0';

	/* buf now contains the port */
	req->header_port = strtoul(buf, NULL, 0);
}

static char *strdup_without_crlf(char *line)
{
	char *p;
	unsigned long bytes = 0;

	for (p = line; *p && (*p != '\r') && (*p != '\n'); p++)
		bytes++;

	if (!(p = malloc(bytes + 1)))
		return NULL;
	
	memcpy(p, line, bytes);
	*(p + bytes) = '\0';

	return p;
}

/* this function reads the request from stdin. It parses it, fills the
 * struct http_request and returns either 0 on success or an approriate
 * error code. The parsing is stopped when the body of the request
 * is reached, so the rest of the server is able to read it via fread.
 */
static int read_request_from_stdin(struct http_request *req)
{
	char line[BUF_SIZ];
	unsigned int line_nr = 0;
	int ret;

	while (fgets(line, sizeof(line), stdin))
	{
		line_nr++;
#ifdef DEBUG
		fprintf(stderr, "examining line %i\n", line_nr);
#endif
		
		if (line[0] == '\r' || line[0] == '\n')
		{
#ifdef DEBUG
			fprintf(stderr, "empty line => header finished\n");
#endif
			break;
		}

		if (line_nr == 1)
		{
			/* save the request line for logging */
			if(!(req->request_line = strdup_without_crlf(line)))
				return HTTP_INTERNAL_SERVER_ERROR;

			if ((ret = parse_method_line(req, line)))
				return ret;
			continue;
		}

		if (!ncmp(line, "host:", 4))
		{
			get_header_host(req, line);
			continue;
		}

		if (!ncmp(line, "content-type:", 13))
		{
			req->content_type = get_header_value(line);
			continue;
		}

		if (!ncmp(line, "user-agent:", 11))
		{
			req->agent = get_header_value(line);
			continue;
		}

		if (!ncmp(line, "referer:", 8))
		{
			req->referer = get_header_value(line);
			continue;
		}

		if (!ncmp(line, "content-length:", 15))
		{
			req->content_length = get_content_length(line);
			continue;
		}

		if (!ncmp(line, "if-modified-since:", 18))
		{
			req->if_modified_since = get_mod_date(line);
			req->if_flags |= IF_MODIFIED_SINCE;
			continue;
		}

		if (!ncmp(line, "if-unmodified-since:", 20))
		{
			req->if_unmodified_since = get_mod_date(line);
			req->if_flags |= IF_UNMODIFIED_SINCE;
			continue;
		}

		if (!ncmp(line, "connection:", 11))
		{
			req->keep_alive = get_keep_alive_state(line);
			continue;
		}

		if (!ncmp(line, "range:", 6))
		{
			if(!(req->ranges = get_ranges(line)) && out_of_memory)
				return HTTP_INTERNAL_SERVER_ERROR;
			continue;
		}

		/* RFC 2616, 14.20: 
		 * --------------------------
		 * A server that does not understand or is unable to comply
		 * with any of the expectation values in the Expect field of
		 * a request MUST respond with appropriate error status.
		 * --------------------------
		 * so we do.
		 */
		if (!ncmp(line, "expect:", 7))
			return HTTP_NOT_IMPLEMENTED;
	}

	return 0;
}

#ifdef DEBUG
static int print_out_request(const struct http_request *req)
{
	struct http_ranges *r;

	printf("===== print_out_request(%p) =====\n", req);

	printf("we are speaking HTTP/%i.%i\n", req->http_major, req->http_minor);
	printf("URI: <%s>\n", req->uri);
	printf("that's <%s> on %s, port %i\n", req->path, req->uri_host,
			req->uri_port);
	printf("additional data: <%s>\n", req->uri_data);
	printf("client says host is %s, port is %i\n", req->header_host,
			req->header_port);

	printf("Content-Length is %lu\n", (unsigned long)req->content_length);

	if (req->if_flags & IF_MODIFIED_SINCE)
		printf("if modified since %s",
				asctime(gmtime(&req->if_modified_since)));

	if (req->if_flags & IF_UNMODIFIED_SINCE)
		printf("if unmodified since %s",
				asctime(gmtime(&req->if_unmodified_since)));

	printf("Content-Type is %s\n", req->content_type);

	printf("Agent is %s\n", req->agent);

	printf("Keep-Alive: %i\n", req->keep_alive);

	printf("Method is: %i\n", req->method);

	if (req->ranges)
	{
		printf("Ranges:\n");
		
		for (r = req->ranges; r; r = r->next)
			printf("\t%lu-%lu (%i)\n", (unsigned long) r->start,
					(unsigned long) r->end, r->valid);
	}
	else
		printf("no Ranges.\n");

	printf("Referer: %s\n", req->referer);
	
	puts("=========================================");

	return 0;
}
#endif

/* do some sanity checks, insert things not filled out (e.g.
 * Connection:)
 */
static int sanity_checks(struct http_request *req,
		const struct server_config *conf)
{
	/* HTTP/1.0: default is no Keep-Alive
	 * HTTP/1.1: default is Keep-Alive
	 * other: invalid version
	 */
	if (req->http_major == 1)
	{
		switch(req->http_minor)
		{
			case 0: if (req->keep_alive == KEEP_ALIVE_NONE)
						req->keep_alive = KEEP_ALIVE_NO;
					break;
			case 1: if (req->keep_alive == KEEP_ALIVE_NONE)
						req->keep_alive = KEEP_ALIVE_YES;
					break;
			default:
					return HTTP_VERSION_NOT_SUPPORTED;
		}
	}
	else
		return HTTP_VERSION_NOT_SUPPORTED;

	if (req->http_major > 1)
		return HTTP_VERSION_NOT_SUPPORTED;

	/* check whether version too high for us, if so we trim it
	 * (only useful if HTTP_VERSION_MINOR is changed from 1
	 * to 0)
	 */
	if ((req->http_major >= HTTP_VERSION_MAJOR) &&
		(req->http_minor > HTTP_VERSION_MINOR))
	{
		req->http_major = HTTP_VERSION_MAJOR;
		req->http_minor = HTTP_VERSION_MINOR;
	}

	/* from now on we can answer in the client's HTTP version */
	req->http_version_checked = 1;

	/* if still no keep alive set, it is no */
	if (req->keep_alive == KEEP_ALIVE_NONE)
		req->keep_alive = KEEP_ALIVE_NO;

	/* no file specified? set default index file */
	if (!cmp(req->path, "/"))
	{
		free(req->path);
		req->path = strdup(conf->index);
	}

	/* HTTP/1.1 demand a Host: line */
	if ((req->http_minor == 1) && !req->header_host)
		return HTTP_BAD_REQUEST;

	return 0;
}

/* we use heavily strdup and malloc for the fields in http_request{},
 * so it is a very good idea to free them :-)
 */

static void free_request(const struct http_request *req)
{
	free(req->request_line);
	free(req->uri);
	free(req->uri_host);
	free(req->header_host);
	free(req->path);
	/* remember: uri_data is a pointer somewhere into uri */
	free(req->content_type);
	free(req->agent);
	free(req->referer);

	if (req->ranges)
		free_ranges(req->ranges);
}


/* the main loop reads the requests from stdin, parses it and then
 * tries to create a proper response. The requests will be read
 * line by line via fgets(), the rest of the request (the body if
 * there is one) is left in the buffer and read via fread (we know
 * its length from the Content-Length field in the header). If the
 * user provides bullshit in the header, well, the fread will fuck
 * up the rest of the piped requests and eventually block the
 * server (if there is not enough data in stdin, we loop, and if
 * there is nothing more provided... stupid).
 *
 * I will think later about this problem. If we would use basic I/O
 * we could use select, but I know the problems one gets when mixing
 * low I/O and high I/O, so we don't. Err, using generally low level
 * I/O is also a bad idea, because it makes things complicated. We
 * are using inetd, so we can consume a little bit CPU power by the
 * comfort of std* functions, right?
 *
 * Perhaps we can set an alarm(2) with a long period of time and
 * just kill the process if there is no data for, let's say 15
 * minutes? Yeah, keep-alive, but who cares?
 */
int main_loop(const struct server_config *conf)
{
	struct http_request req;
	int ret, keep_alive = 1;
	
	while (keep_alive)
	{
		memset(&req, 0, sizeof(req));

		/* read request from stdin, parse it into req */
		if ((ret = read_request_from_stdin(&req)))
		{
			http_send_error(conf, &req, ret);
			return 101;
		}

#ifdef DEBUG
		fprintf(stderr, "now doing some sanity checks\n");
#endif

		/* do sanity checks, after this point the request
		 * is syntactically OK and really IS HTTP.
		 */
		if ((ret = sanity_checks(&req, conf)))
		{
			http_send_error(conf, &req, ret);
			return 102;
		}

#ifdef DEBUG
		/* view the values in req */
		print_out_request(&req);
#endif

		/* good idea: do something with the request :-) */
		ret = send_response(conf, &req);

		/* do we (still) use Keep-Alive? If so, loop once again,
		 * if not, leave the loop next time.
		 */
		if (req.keep_alive != KEEP_ALIVE_YES)
			keep_alive = 0;

		/* just to be sure */
		fflush(stdout);

		logfile_entry(conf, &req, ret);

		/* we used lots of strdups, good idea to free them */
		free_request(&req);
	}
	
	return 100;
}

