/*
 * httpd.h - the central header file for httpd
 *
 * Copyright (c) 2003
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef HTTPD_H
#define HTTPD_H

#define BUF_SIZ	4096
#define VERSION_MAJOR	0
#define VERSION_MINOR	1
#define VERSION_STRING	"rc-1"
#define SERVER_NAME		"httpd"

#define HTTP_VERSION_MAJOR	1
#define HTTP_VERSION_MINOR	1

#define DEFAULT_WWW_DIR	"/var/www/"
#define DEFAULT_INDEX	"index.html"
#define DEFAULT_CONFIG	"/usr/local/etc/httpd.conf"
#define DEFAULT_CONTENT_TYPE	"text/plain"

#define OPTION_DISABLE	1
#define OPTION_ENABLE	2

struct server_config {
	const char	*www_dir;
	uid_t		uid;
	gid_t		gid;
	const char	*index;
	const char	*error_file_path;
	const char	*default_ct;
	char		dotfiles;
	char		extlog;
	char		dirlisting;
};


#define RANGE_START	1
#define RANGE_END	2

struct http_ranges {
	size_t				start, end;
	unsigned			valid;
	struct http_ranges	*next;
};

enum keep_alive_states {KEEP_ALIVE_NONE, KEEP_ALIVE_NO, KEEP_ALIVE_YES};

/* NOTE: POST will be implemented later, not in 0.1.0.
 * You want to add the support at parse_method_line() in request.c
 * in order to stop httpd throwing HTTP 501.
 * The right position to add the functionality is response.c, I think
 * you should start somewhere in send_respone(), there is the stuff
 * for GET and HEAD. Also add the CGI support here.
 */
enum http_methods {METHOD_NONE, METHOD_GET, METHOD_HEAD, METHOD_POST};

#define	IF_MODIFIED_SINCE	1
#define IF_UNMODIFIED_SINCE	2

#define HTTP_OK						200
#define HTTP_PARTIAL_CONTENT		206
#define HTTP_MOVED_PERMANENTLY		301
#define HTTP_NOT_MODIFIED			304
#define HTTP_MOVED_TEMPORARILY		307
#define HTTP_BAD_REQUEST			400
#define HTTP_FORBIDDEN				403
#define HTTP_NOT_FOUND				404
#define HTTP_PRECONDITION_FAILED	412
#define HTTP_RANGE_NOT_SATISFIABLE	416
#define HTTP_INTERNAL_SERVER_ERROR	500
#define HTTP_NOT_IMPLEMENTED		501
#define HTTP_VERSION_NOT_SUPPORTED	505

/* implementation reserved space */
#define STAT_IS_DIRECTORY			600
#define DIR_HAS_INDEX				601

struct http_request {
	unsigned				http_major, http_minor, http_version_checked;
	char					*uri;
	char					*uri_host;
	char					*header_host;
	unsigned short			uri_port;
	unsigned short			header_port;
	char					*path;
	char					*uri_data;
	size_t					content_length;
	unsigned				if_flags;
	time_t					if_modified_since;
	time_t					if_unmodified_since;
	char					*content_type;
	char					*agent;
	enum keep_alive_states	keep_alive;
	enum http_methods		method;
	struct http_ranges		*ranges;
	char					*referer;
	size_t					bytes_sent;
	char					*request_line;
};

struct http_response {
	unsigned				http_major, http_minor;
	char					*path;
	const char				*content_type;
	size_t					content_length;
	time_t					last_modified;
	enum keep_alive_states	keep_alive;
	enum http_methods		method;
	struct http_ranges		*ranges;
	size_t					content_sent;
};

struct chunk_list {
	char				*data, *header;
	size_t				data_len, header_len;
	char				*boundary;
	struct chunk_list	*next;
};

/* config.c */
int config_load_file(struct server_config*, const char*, int);
#ifdef DEBUG
void config_print(const struct server_config*);
#endif
int config_set_defaults(struct server_config*);
int config_set_ids(struct server_config*, const char*, const char*);

/* get_head.c */
int http_get_head(struct http_response*);

/* misc.c */
int cmp(const char*, const char*);
const char *get_content_type(const char*, const struct server_config*);
const char *http_datestring(const time_t);
int http_send_dirlisting(const struct http_request*,
		struct http_response*);
unsigned long http_send_error(const struct server_config*,
		struct http_request*, int);
const char *http_serverstring(void);
int logfile_entry(const struct server_config*,
		const struct http_request*, int);
int ncmp(const char*, const char*, size_t);
const char *response_message_short(int);

/* request.c */
int main_loop(const struct server_config*);

/* response.c */
int send_response(const struct server_config*, struct http_request*);

#ifndef min
#define min(a,b) ((a) <= (b) ? (a) : (b))
#endif

#endif

